unit jpmPrint;
{Copyright 2000-2001, J. Peter Mugaas}
interface

{$I JPM.INC}

uses Graphics, Printers, SysUtils, Windows;

{$IFNDEF D4andAbove}
type
     LongWord = Cardinal;
     {$ENDIF}

type
     {: This type is used for specifying Horizontal Text alignment.  jpmHorizLeft means left alignment. JPMHorizCenter means to center the text.  JPMHorizRight means to align text to the right}
     TJPMAlign = (jpmHorizLeft, JPMHorizCenter, JPMHorizRight);
     {: This is a simple printing object that prints text in a series of method calls in a sequence along with optional headers and footers.  Pages are handled automatically}
     TJPMSimplePrint = class (TObject)

     private
          FTextAlign: TJPMAlign;
          FHeader, FFooter: string;
          FHeaderAlign, FFooterAlign: TJPMAlign;
          FHeaderFont, FFooterFont, FTextFont: TFont;
          FPrinter: TPrinter;
          FLine: Integer;
          fMargin: TPoint;
          FIdentifierValueSeparator: string;
          procedure SetPrinterFont (Font: TFont);
          function GetAPIAlignFlag (Align: TJPMAlign): DWord;
          procedure GetMargenOffset;
          function CalcHeaderFooter (const AHeader: string): string;
          function TextHeight (const AText: string): Integer;
          procedure RawPrintLine (const AText: string; const Align: TJPMAlign);
          function FooterHeaderHeight (const AFooter: string): Integer;
          procedure PrintFooter (const AFooter: string);
          procedure PrintHeader (const AHeader: string);
          procedure SetFooterFont (const Value: TFont);
          procedure SetHeaderFont (const Value: TFont);
          procedure SetTextFont (const Value: TFont);
          procedure NewPage;
     public
          {: This constructor creates a new TJPMSimplePrint object.  APRinter is the printer to use for the printing}
          constructor Create (APrinter: TPrinter);
          {: This destructor destroys the TJPMSimplePrint object}
          destructor Destroy; override;
          {: This starts a print job}
          procedure BeginPrint;
          {: This ends a print job}
          procedure EndPrint;
          {: This prints a line of text to the printer with word wrapping}
          procedure PrintLine (const Text: string);
          {: This prints the identifier followed by the string value}
          procedure PrintStrVal (const Identifier, Value: string);
          {: This prints the identifier followed by the boolean value}
          procedure PrintBoolVal (const Identifier: string; const Value: Boolean);
          {: This prints the identifier followed by the number}
          procedure PrintNumVal (const Identifier: string; const Value: LongWord);
          {: This is a header which will be printed on every page.  If you want to print the page number, just include the string "#p" where the number should be.  Leave this blank if no header is desired.}
          property Header: string read FHeader write FHeader;
          {: This is a footer which will be printed on every page.  If you want to print the page number, just include the string "#p" where the number should be.  Leave this blank if no footer is desired.}
          property Footer: string read FFooter write FFooter;
          {: This is the horizontal alignment for the header}
          property HeaderAlign: TJPMAlign read FHeaderAlign write FHeaderAlign;
          {: This is the horizontal alignment for the footer}
          property FooterAlign: TJPMAlign read FFooterAlign write FFooterAlign;
          {: This is the horizontal alignment for the text you print with the print methods}
          property TextAlign: TJPMAlign read FTextAlign write FTextAlign;
          {: This is the Font for the header}
          property HeaderFont: TFont read FHeaderFont write SetHeaderFont;
          {: This is the Font for the footer}
          property FooterFont: TFont read FFooterFont write SetFooterFont;
          {: This is the Font for the fext you print with the print methods}
          property TextFont: TFont read FTextFont write SetTextFont;
          {: This is the TPrinter object you specified with the constructor}
          property Printer: TPrinter read FPrinter;
          {: This property specifies a separator between an identifier and a value }
          property IdentifierValueSeparator: string read FIdentifierValueSeparator write FIdentifierValueSeparator;
     end;

implementation

{ TJPMSimplePrint }

procedure TJPMSimplePrint.BeginPrint;
begin
     GetMargenOffset;
     FPrinter.BeginDoc;
     PrintHeader (FHeader);
end;

function TJPMSimplePrint.CalcHeaderFooter (const AHeader: string): string;
begin
     Result := StringReplace (AHeader, '#p', IntToStr (FPrinter.PageNumber), [rfReplaceAll]);
end;

constructor TJPMSimplePrint.Create (APrinter: TPrinter);
begin
     inherited Create;
     IdentifierValueSeparator := ': ';
     FPrinter := APrinter;
     FHeaderFont := TFont.Create;
     FFooterFont := TFont.Create;
     FTextFont := TFont.Create;
end;

destructor TJPMSimplePrint.Destroy;
begin
     FHeaderFont.Free;
     FFooterFont.Free;
     FTextFont.Free;
     inherited Destroy;
end;

procedure TJPMSimplePrint.EndPrint;
begin
     PrintFooter (FFooter);
     FPrinter.EndDoc;
end;

function TJPMSimplePrint.FooterHeaderHeight (
     const AFooter: string): Integer;
var
     s: string;
begin
     s := CalcHeaderFooter (AFooter);
     Result := TextHeight (s);
end;

function TJPMSimplePrint.GetAPIAlignFlag (Align: TJPMAlign): DWord;
begin
     case Align of
          jpmHorizLeft: Result := DT_LEFT;
          JPMHorizCenter: Result := DT_CENTER;
          JPMHorizRight: Result := DT_RIGHT;
     else
          Result := DT_LEFT;
     end;
end;

procedure TJPMSimplePrint.GetMargenOffset;
begin
     fMargin.x := GetDeviceCaps (FPrinter.Handle, PHYSICALOFFSETX);
     fMargin.y := GetDeviceCaps (FPrinter.Handle, PHYSICALOFFSETY);
end;

procedure TJPMSimplePrint.NewPage;
begin
     FPrinter.NewPage;
     FLine := fMargin.Y;
end;

procedure TJPMSimplePrint.PrintBoolVal (const Identifier: string;
     const Value: Boolean);
begin
     if Value then
          PrintLine (Identifier + ' False')
     else
          PrintLine (Identifier + ' True');
end;

procedure TJPMSimplePrint.PrintFooter (const AFooter: string);
var
     s: string;
begin
     if Length (AFooter) > 0 then
     begin
          SetPrinterFont (FFooterFont);
          FLine := (FPrinter.PageHeight - FooterHeaderHeight (' ' + #10#13 + Footer));
          s := CalcHeaderFooter (AFooter);
          RawPrintLine (#10#13 + s, FooterAlign);
     end;
end;

procedure TJPMSimplePrint.PrintHeader (const AHeader: string);
var
     s: string;
begin
     if Length (AHeader) > 0 then
     begin
          SetPrinterFont (HeaderFont);
          FLine := fMargin.Y;
          s := CalcHeaderFooter (AHeader);
          RawPrintLine (s + #10#13 + ' ', HeaderAlign);
     end;
end;

procedure TJPMSimplePrint.PrintLine (const Text: string);
var
     Height: Integer;
begin
     {we have print out the header for the first page}
     if (FPrinter.PageNumber = 1) and (FLine = fMargin.y) then
          PrintHeader (FHeader);
     {text Font}
     SetPrinterFont (FTextFont);
     {This is so this routine will skip a line}
     Height := FPrinter.Canvas.TextHeight (Text);
     if (FLine + Height >= (FPrinter.PageHeight - FooterHeaderHeight (' ' + #10#13 + Footer))) then
     begin
          {Move to next page after printing the footer and disregard the line if it is blank}
          PrintFooter (FFooter);
          NewPage;
          FLine := fMargin.y;
          PrintHeader (FHeader);
          {Just in case the header Font is different than the text Font}
          SetPrinterFont (FTextFont);
          RawPrintLine (Text, FTextAlign);
     end
     else
     begin
          if Length (Text) = 0 then
               RawPrintLine (' ', TextAlign)
          else
               {we only print a blank line if there is no new page}
               RawPrintLine (Text, TextAlign);
     end;
end;

procedure TJPMSimplePrint.PrintNumVal (const Identifier: string;
     const Value: Cardinal);
begin
     if Value <> 0 then
          PrintStrVal (Identifier, IntToStr (Value));
end;

procedure TJPMSimplePrint.PrintStrVal (const Identifier, Value: string);
begin
     if Length (Value) <> 0 then
          PrintLine (Identifier + IdentifierValueSeparator + Value);
end;

procedure TJPMSimplePrint.RawPrintLine (const AText: string;
     const Align: TJPMAlign);
var
     Height: Integer;
     Rect: TRect;
begin
     Rect := FPrinter.Canvas.ClipRect;
     Rect.Top := FLine;
     DrawText (FPrinter.Canvas.Handle, PChar (AText), Length (AText), rect, DT_WORDBREAK or GetAPIAlignFlag (Align));
     Height := TextHeight (AText);
     FLine := FLine + Height;
end;

procedure TJPMSimplePrint.SetFooterFont (const Value: TFont);
begin
     FFooterFont.Assign (Value);
end;

procedure TJPMSimplePrint.SetHeaderFont (const Value: TFont);
begin
     FHeaderFont.Assign (Value);
end;

procedure TJPMSimplePrint.SetPrinterFont (Font: TFont);
begin
     FPrinter.Canvas.Font.Charset := Font.Charset;
     FPrinter.Canvas.Font.Color := Font.Color;
     FPrinter.Canvas.Font.Name := Font.Name;
     FPrinter.Canvas.Font.Pitch := Font.Pitch;
     FPrinter.Canvas.Font.Size := Font.Size;
     FPrinter.Canvas.Font.Style := Font.Style;
end;

procedure TJPMSimplePrint.SetTextFont (const Value: TFont);
begin
     FTextFont.Assign (Value);
end;

function TJPMSimplePrint.TextHeight (const AText: string): Integer;
var
     Rect: TRect;
begin
     Rect := FPrinter.Canvas.ClipRect;
     Result := DrawText (FPrinter.Canvas.Handle, PChar (AText), Length (AText), Rect, DT_CALCRECT or DT_WORDBREAK or GetAPIAlignFlag (TextAlign));
end;

end.
