{ ##
  @PROJECT_NAME             Windows shell folders routines and components.
  @PROJECT_DESC             A set of routines, classes and components for
                            processing Windows shell folders. Includes:
                            + Component that provides information about a
                              specified special shell folder.
                            + Class that enumerates all the special shell
                              folders.
                            + Routines to manipulate special shell folder
                              information.
                            + Component that encapsulates the Windows Browse for
                              Folder dialog box.
  @FILE                     PJShellFoldersDsgn.pas
  @COMMENTS                 Component registration code and property editor for
                            TPJBrowseDialog and TPJSpecialFolderInfo components
                            defined in PJShellFolders.pas.
  @LEGAL_NOTICE             The source code and any help files can be freely
                            distributed on a not-for-profit basis providing
                            that:
                            + the source code is not altered and
                            + these comments are not removed from the source
                              file.\
                            By not-for-profit I mean that you may recover out of
                            pocket expenses incurred in distributing the code,
                            but should not make a profit from this. If you
                            discover any bugs in this implementation, or if you
                            have any update suggestions, please contact the
                            author. Please do modify the code for you own use.
                            I'd like to see any changes you make - I could
                            incorporate them into future versions. Please notify
                            the author of changes. This software is provided as
                            is - no warranty is given as to its suitability for
                            any purposes to which you may wish to put it.
  @EMAIL                    peter.johnson@openlink.org
  @WEBSITE                  http://www.delphidabbler.com/
  @AUTHOR                   Peter D Johnson, Llanarth, Ceredigion, Wales, UK.
  @COPYRIGHT                Copyright  2001-2003, P.D.Johnson, Llanarth,
                            Ceredigion, Wales UK.
  @HISTORY(
    @REVISION(
      @VERSION              1.0
      @DATE                 15/06/2003
      @COMMENTS             Original version.
    )
  )
}


unit PJShellFoldersDsgn;


interface


uses
  {$IFDEF VER140}
    {$DEFINE DELPHI6ANDUP}
  {$ENDIF}
  {$IFDEF VER150}
    {$DEFINE DELPHI6ANDUP}
    {$DEFINE DELPHI7ANDUP}
  {$ENDIF}

  // Delphi
  Classes,
  {$IFDEF DELPHI6ANDUP}
    DesignIntf, DesignEditors;
  {$ELSE}
    DsgnIntf;
  {$ENDIF}


type
  {
  TPJFolderIDPE:
    Property editor for special folder ID properties. Permits choice of
    predefined special folders IDs.
  }
  TPJFolderIDPE = class(TPropertyEditor)
  public
    function GetAttributes: TPropertyAttributes; override;
      {Tell object inspector that a list of read only values should be
      displayed}
    procedure GetValues(Proc: TGetStrProc); override;
      {Returns all available folder ids for display in drop-down list}
    procedure SetValue(const Value: string); override;
      {Returns actual numeric value for given folder id constant}
    function GetValue: string; override;
      {Returns the name of the folder id constant for the given identifier
      value}
  end;


procedure Register;
  {Delphi component library registration routine}


implementation


uses
  // Delphi
  SysUtils, ShlObj,
  // Project
  PJShellFolders;


// -----------------------------------------------------------------------------
// TPJFolderIDPE
// -----------------------------------------------------------------------------

function TPJFolderIDPE.GetAttributes: TPropertyAttributes;
  {Tell object inspector that a list of read only values should be displayed}
begin
  Result := [paValueList, paReadOnly];
end;

function TPJFolderIDPE.GetValue: string;
  {Returns the name of the folder id constant for the given identifier value}
begin
  Result := SpecialFolderIDToStr(GetOrdValue);
end;

procedure TPJFolderIDPE.GetValues(Proc: TGetStrProc);
  {Returns all available folder ids for display in drop-down list}
var
  Enum: IPJSpecialFolderEnum; // enumerator for folder ids
begin
  Enum := TPJSpecialFolderEnum.Create;
  Enum.Init;
  while not Enum.AtEnd do
    Proc(SpecialFolderIdToStr(Enum.Next));
end;

procedure TPJFolderIDPE.SetValue(const Value: string);
  {Returns actual numeric value for given folder id constant}
begin
  SetOrdValue(StrToSpecialFolderID(Value));
end;


// -----------------------------------------------------------------------------
// Delphi registration routine
// -----------------------------------------------------------------------------

procedure Register;
  {Registers components and property editor with Delphi}
begin
  // Register the unit's components
  RegisterComponents('DelphiDabbler', [TPJSpecialFolderInfo, TPJBrowseDialog]);

  // Register special folder id property editor:
  // .. for TPJSpecialFolderInfo.FolderID
  RegisterPropertyEditor(
    TypeInfo(Integer),        // type information about property we edit
    TPJSpecialFolderInfo,     // work with component of this type
    'FolderID',               // only for use with this property
    TPJFolderIDPE);           // property editor class
  // .. for TPJBrowseDlg.RootFolderID
  RegisterPropertyEditor(
    TypeInfo(Integer),        // type information about property we edit
    TPJBrowseDialog,          // work with component of this type
    'RootFolderID',           // only for use with this property
    TPJFolderIDPE);           // property editor class
end;

end.
