{*******************************************************}
{                                                       }
{         Copyright (c) FINANCES Ltd                    }
{         2003   ver. 1                                 }
{                                                       }
{         www.finances-ltd.narod.ru                     }
{         finances-ltd@narod.ru                         }
{                                                       }
{*******************************************************}

// Improvide version of TRessourceStore (Alexandre GUILLIEN)

{
 Component that allows the developper to add compressed custom resources in its
 project through the delphi IDE.
 
 Compress with ZLib or LHA (see below)
 
 property
   StoreMode
   DataSize
   FileSize
   Name
   FileName

Methods
   SaveToStream
   LoadFromStream
   SaveToFile
   LoadFromFile
   SaveToStrings
   GetAsStream
}


unit zResStorage;

{.$DEFINE USEZIP}
{$DEFINE USELHA}

interface

uses
  SysUtils, Classes;

type
  ERessourceStoreError = class(Exception);
  TPackedRessources = class;
  TPackedItem = class;

  TzResStorage = class(TComponent)
  private
    FRessources: TPackedRessources;
    procedure SetRessources(Ressources: TPackedRessources);
    function GetSize: Int64;
    procedure SetSize(NewSize: Int64);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    {}
    procedure SaveToPath(PathName: string);
    {}
    function ItemIndex(ItemName: string): Integer;
    function GetRes(ResName: string): TPackedItem; overload;
    function GetRes(Index: Integer): TPackedItem; overload;
    function ResExists(ItemName: string): Boolean;
    function SaveToStrings(Dest: TStrings; ItemName: string): Boolean;
    function SaveToFile(FileName: string; ItemName: string): Boolean;
    function SaveToStream(Stream: TStream; ItemName: string): Boolean;
  published
    property Items: TPackedRessources read FRessources write SetRessources;
    property Size: Int64 read GetSize write SetSize;
  end;

  TPackedRessources = class(TCollection)
  private
    FOwner: TPersistent;
    function GetItem(Idx: Integer): TPackedItem;
    procedure SetItem(Idx: Integer; Item: TPackedItem);
  protected
    function GetOwner: TPersistent; override;
    {}
    property MyOwner: TPersistent read FOwner write FOwner;
  public
    constructor Create;
    property Items[Idx: Integer]: TPackedItem read GetItem write SetItem; default;
  end;

  TStoreMode = (smStore{$IFDEF USEZIP}, smZipLow, smZipMed, smZipMax{$ENDIF});

  TPackedItem = class(TCollectionItem)
  private
    FData: TMemoryStream;
    FStoreMode: TStoreMode;
    FName: string;
    FFileName: TFileName;
    FFileSize: Int64;
    FUnCompressedStream: TMemoryStream;
    procedure SetStoreMode(StoreMode: TStoreMode);
    procedure ReadData(Stream: TStream);
    procedure WriteData(Stream: TStream);
    function GetSize: Integer;
    procedure SetSize(NewSize: Integer);
    procedure SetFileSize(Value: Int64);
    procedure SetName(Name: string);
    Procedure SetFileName( Value: TFileName );
  protected
    function GetDisplayName: string; override;
    procedure DefineProperties(Filer: TFiler); override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    {}
    procedure SaveToStream(Stream: TStream);
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToFile(FileName: string);
    procedure LoadFromFile(FileName: string);
    procedure SaveToStrings(Dest: TStrings);
    {}
    function GetAsStream: TMemoryStream;
    {}
    property Data: TMemoryStream read GetAsStream;
  published
    property StoreMode: TStoreMode read FStoreMode write SetStoreMode {$IFDEF USEZIP}default smZipMax{$ENDIF};
    property DataSize: Integer read GetSize write SetSize;
    property FileSize: Int64 read FFileSize write SetFileSize;
    property Name: string read FName write SetName;
    property FileName: TFileName Read FFileName Write SetFileName;
  end;

  procedure Register;

implementation

{$IFDEF USEZIP}
uses ZLib;
{$ELSE}
{$IFDEF USELHA}
{$ENDIF}
uses LHARoutines;
{$ENDIF}

const
  msgRessourceNotFound = 'Ressource "%s" not found';
  msgNameAlreadyExists = 'Item name "%s" already exists !';

{ ****************************** TzResStorage ****************************** }

constructor TzResStorage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FRessources:= TPackedRessources.Create;
  FRessources.MyOwner:= Self;
end;

destructor TzResStorage.Destroy;
begin
  FRessources.Free;
  inherited Destroy;
end;

function TzResStorage.ItemIndex(ItemName: string): Integer;
var i: Integer;
begin
  Result:= -1;
  i:= 0;
  while (i < FRessources.Count) and (CompareText(ItemName, TPackedItem(FRessources.Items[i]).Name) <> 0) do
    Inc(i);
  if i < FRessources.Count then
    Result:= i;
end;

function TzResStorage.ResExists(ItemName: string): Boolean;
begin
  Result := ItemIndex(ItemName) <> -1;
end;

function TzResStorage.GetRes(ResName: string): TPackedItem;
var Idx: Integer;
begin
  Idx:= ItemIndex(ResName);
  if Idx <> -1 then
    Result:= TPackedItem(FRessources.Items[Idx])
  else raise ERessourceStoreError.CreateFmt(msgRessourceNotFound, [ResName]);
end;

function TzResStorage.GetRes(Index: Integer): TPackedItem;
begin
  if Index in [0..Items.Count] then
    Result:= Items[Index]
  else Result := nil;
end;

function TzResStorage.GetSize: Int64;
var i: Integer;
begin
  Result:= 0;
  for i:= 0 to FRessources.Count - 1 do
    Result:= Result + TPackedItem(FRessources.Items[i]).DataSize;
end;

procedure TzResStorage.SetSize(NewSize: Int64); // ReadOnly property
begin end;

procedure TzResStorage.SetRessources(Ressources: TPackedRessources);
begin
  FRessources.Assign(Ressources);
end;

procedure TzResStorage.SaveToPath(PathName: string);
var i: Integer;
begin
  if (PathName <> '') and (PathName[Length(PathName)] <> '\') then
    PathName:= PathName + '\';
  for i:= 0 to Items.Count - 1 do
    Items[i].SaveToFile(PathName + Items[i].Name);
end;

function TzResStorage.SaveToStrings(Dest: TStrings; ItemName: string): Boolean;
begin
  Result := ResExists(ItemName);
  Dest.Clear;
  if Result then
    GetRes(ItemName).SaveToStrings(Dest)
  else Dest.Clear;
end;

function TzResStorage.SaveToFile(FileName: string; ItemName: string): Boolean;
begin
  Result := ResExists(ItemName);
  if Result then
    GetRes(ItemName).SaveToFile(FileName);
end;

function TzResStorage.SaveToStream(Stream: TStream; ItemName: string): Boolean;
begin
  Result := ResExists(ItemName);
  if Result then
    GetRes(ItemName).SaveToStream(Stream);
end;

{ ******************************** TPackedItem ******************************** }

constructor TPackedItem.Create(Collection: TCollection);
begin
  {$IFDEF USEZIP}FStoreMode:= smZipMax;{$ENDIF}
  inherited Create(Collection);
  FData := TMemoryStream.Create;
  FUnCompressedStream := TMemoryStream.Create;
end;

destructor TPackedItem.Destroy;
begin
  FData.Free;
  FData:= nil;
  FUnCompressedStream.Free;
  FUnCompressedStream := nil;
  inherited Destroy;
end;

procedure TPackedItem.Assign(Source: TPersistent);
begin
  if Source is TPackedItem then
  begin
    with TPackedItem(Source) do
    begin
      Self.Name:= Name;
      Self.FData.Clear;
      Self.StoreMode:= StoreMode;
      Self.FData.LoadFromStream(FData);
    end;
  end else AssignTo(Source);
end;

{$IFDEF USEZIP}
  procedure TPackedItem.SaveToStream(Stream: TStream);
  var S: TDecompressionStream;
      L: Integer;
  begin
    FData.Position:= 0;
    if (StoreMode <> smStore) and (FData.Size > 0) then
    begin
      S:= TDecompressionStream.Create(FData);
      try
        S.Read(L, SizeOf(Integer));
        if L <> 0 then
          Stream.CopyFrom(S, L);
      finally
        S.Free;
      end;
    end else
      Stream.CopyFrom(FData, FData.Size);
  end;

  procedure TPackedItem.LoadFromStream(Stream: TStream);
  var S: TCompressionStream;
      L: TCompressionLevel;
      Tmp: Integer;
  begin
    case StoreMode of
      smZipLow: L:= clFastest;
      smZipMed: L:= clDefault;
      smZipMax: L:= clMax
      else L:= clNone;
    end;
    Stream.Position:= 0;
    FData.Clear;
    if Stream.Size = 0 then
      Exit;
    if L <> clNone then
    begin
      S:= TCompressionStream.Create(L, FData);
      try
        Tmp:= Stream.Size;
        S.Write(Tmp, SizeOf(Integer));
        S.CopyFrom(Stream, Tmp);
      finally
        S.Free;
      end;
    end else
      FData.CopyFrom(Stream, 0);
  end;
{$ELSE}
  {$IFDEF USELHA}
    procedure TPackedItem.SaveToStream(Stream: TStream);
    begin
      LHAExpand(FData, Stream);
    end;

    procedure TPackedItem.LoadFromStream(Stream: TStream);
    begin
      FData.Clear;
      FFileSize := Stream.Size;
      LHACompress(Stream, fData);
    end;
  {$ELSE}
    procedure TPackedItem.SaveToStream(Stream: TStream);
    begin
      FData.Position:= 0;
      Stream.CopyFrom(FData, 0);
    end;

    procedure TPackedItem.LoadFromStream(Stream: TStream);
    begin
      FData.Clear;
      FData.CopyFrom(Stream, 0);
    end;
  {$ENDIF}
{$ENDIF}

procedure TPackedItem.SaveToFile(FileName: string);
var F: TFileStream;
begin
  F:= TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(F);
  finally
    F.Free;
  end;
end;

procedure TPackedItem.LoadFromFile(FileName: string);

  procedure SetUniqueName(aName: string);
  var i: Integer;
      Base: string;
  begin
    try
      Name:= aName;
    except
      Base:= aName;
      i:= 0;
      while Name <> aName do begin
        Inc(i);
        aName:= Base + IntToStr(i);
        try
          Name:= aName;
        except
        end;
      end;
    end;
  end;

var F: TFileStream;
begin
  F:= TFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
  try
    FFileName := FileName;
    SetUniqueName(ExtractFileName(FileName));
    LoadFromStream(F);
  finally
    F.Free;
  end;
end;

function TPackedItem.GetAsStream: TMemoryStream;
begin
  SetStoreMode(smStore);
  {$IFDEF USEZIP}
    if FUnCompressedStream.Size = 0 then
      SaveToStream(FUnCompressedStream);
    Result:= FUnCompressedStream;
  {$ELSE}
    {$IFDEF USELHA}
      if FUnCompressedStream.Size = 0 then
        SaveToStream(FUnCompressedStream);
      Result:= FUnCompressedStream;
    {$ELSE}
      FData.Position:= 0;
      Result:= FData;
    {$ENDIF}
  {$ENDIF}
end;

procedure TPackedItem.SetStoreMode(StoreMode: TStoreMode);
{$IFDEF USEZIP}
var M: TMemoryStream;
begin
  if csLoading in TzResStorage(TPackedRessources(GetOwner).GetOwner).ComponentState then
    FStoreMode:= StoreMode
  else if StoreMode <> FStoreMode then
  begin
    M:= TMemoryStream.Create;
    try
      SaveToStream(M);
      FStoreMode:= StoreMode;
      LoadFromStream(M);
    finally
      M.Free;
    end;
  end;
end;
{$ELSE}
begin
  FStoreMode:= smStore;
end;
{$ENDIF}

function TPackedItem.GetDisplayName: string;
begin
  if FData.Size = 0 then
    Result:= ClassName
  else
    Result:= FName;
end;

function TPackedItem.GetSize: Integer;
begin
  Result:= FData.Size;
end;

procedure TPackedItem.SetSize(NewSize: Integer);
begin end;

procedure TPackedItem.SetFileSize(Value: Int64);
begin
  if (csLoading in TzResStorage(TPackedRessources(GetOwner).GetOwner).componentstate) then
    FFileSize := Value;
end;

procedure TPackedItem.SetName(Name: string);
var i: Integer;
begin
  for i:= 0 to Collection.Count - 1 do
  begin
    if (Collection.Items[i].ID <> ID) and (CompareText(Name, TPackedItem(Collection.Items[i]).Name) = 0) then
      raise ERessourceStoreError.CreateFmt(msgNameAlreadyExists, [Name]);
  end;
  FName:= Name;
end;

procedure TPackedItem.ReadData(Stream: TStream);
{$IFNDEF USELHA}
var L: Integer;
{$ENDIF}
begin
  FData.Clear;
  {$IFNDEF USELHA}
  Stream.Read(L, SizeOf(Integer));
  FData.Size:= L; { Allocates what's needed }
  {$ENDIF}
  FData.CopyFrom(Stream, 0);
end;

procedure TPackedItem.WriteData(Stream: TStream);
{$IFNDEF USELHA}
var L: Integer;
{$ENDIF}
begin
  FData.Position:= 0;
  {$IFNDEF USELHA}
  L:= FData.Size;
  Stream.Write(L, SizeOf(Integer));
  {$ENDIF}
  Stream.CopyFrom(FData, 0);
end;

procedure TPackedItem.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('ResData', ReadData, WriteData, Assigned(FData) and (FData.Size > 0));
end;

procedure TPackedItem.SetFileName(Value: TFileName);
begin
  if (csDesigning in TzResStorage(TPackedRessources(GetOwner).GetOwner).ComponentState) then begin
    if FileExists(Value) then begin
      LoadFromFile(Value);
      FFileName := Value;
    end;
  end
  else if (csLoading in TzResStorage(TPackedRessources(GetOwner).GetOwner).ComponentState) then begin
    FFileName := ExtractFileName(Value);
  end;
end;

procedure TPackedItem.SaveToStrings(Dest: TStrings);
var ms: TMemoryStream;
begin
  ms := TMemoryStream.Create;
  try
    SaveToStream(ms);
    ms.Position := 0;
    Dest.LoadFromStream(ms);
  finally
    ms.Free;
  end;
end;

{ ***************************** TPackedRessources ***************************** }

function TPackedRessources.GetOwner: TPersistent;
begin
  Result:= FOwner;
end;

function TPackedRessources.GetItem(Idx: Integer): TPackedItem;
begin
  Result:= TPackedItem(inherited Items[Idx]);
end;

procedure TPackedRessources.SetItem(Idx: Integer; Item: TPackedItem);
begin
  inherited Items[Idx]:= Item;
end;

constructor TPackedRessources.Create;
begin
  FOwner:= nil;
  inherited Create(TPackedItem);
end;


procedure Register;
begin
  RegisterComponents('D5', [TzResStorage]);
end;

end.
